"""In this module you will find Pane extensions to Panel. For example the PandasProfileReport"""
from .code import Code
from .model_viewer import ModelViewer
from .pandas_profile_report import PandasProfileReport



"""This module provides button to share on social media"""
import urllib.parse

import panel as pn
import param

from awesome_panel_extensions.assets import svg_icons

DEFAULT_URL = "https://awesome-panel.org"
DEFAULT_TEXT = "Checkout"
STYLE = """.bk a.button-share-link {
    color: inherit;
    font-style: none;
}
svg.pnx-icon {
    height: 1em;
    width: 1em;
}
"""
if not STYLE in pn.config.raw_css:
    pn.config.raw_css.append(STYLE)


class ShareOnBase(pn.pane.HTML):
    """Base class for implementing ShareOnFacebook, ShareOnLinkedIn links etc.

    - The href property should be overridden
    """

    url = param.String(DEFAULT_URL)
    icon = param.String(svg_icons.EXTERNAL_LINK, doc="A SVG icon")
    text = param.String(DEFAULT_TEXT)
    size = param.Integer(default=1, bounds=(0, 20), doc="The fontsize in em")

    priority = 0
    _rename = dict(pn.pane.HTML._rename, url=None, icon=None, text=None, size=None)

    def __init__(self, **params):
        super().__init__(**params)

        self._update_html_object()

    @property
    def _url_parsed(
        self,
    ):
        return urllib.parse.quote(self.url).replace(
            "/",
            "%2F",
        )

    @property
    def href(
        self,
    ) -> str:
        """The href to goto when clicked

        Override this method in a base class

        Raises:
            NotImplementedError:

        Returns:
            str: A href string
        """
        raise NotImplementedError()

    def to_html(
        self,
    ) -> str:
        """A html string with link and icon tags

        Returns:
            str: A html string with link and icon tags
        """
        return (
            f'<a href="{self.href}" class="button-share-link" style="font-size: {self.size}em" '
            f'target="_blank">{self.icon}</a>'
        )

    @param.depends("url", "icon", "size", watch=True)
    def _update_html_object(
        self,
    ):
        """A HTML pane with the a link and icon"""
        self.object = self.to_html()


class ShareOnFacebook(ShareOnBase):
    """A Share on Facebook button"""

    icon = param.String(svg_icons.FACEBOOK)

    @property
    def href(
        self,
    ):
        return f"https://www.facebook.com/sharer/sharer.php?u={self._url_parsed}"


class ShareOnLinkedIn(ShareOnBase):
    """A Share on LinkedIn button"""

    icon = param.String(svg_icons.LINKED_IN)

    @property
    def href(
        self,
    ):
        return (
            f"http://www.linkedin.com/shareArticle?mini=true&url={self._url_parsed}"
            f"&title={self.text}"
        )


class ShareOnTwitter(ShareOnBase):
    """A Share on Twitter button"""

    icon = param.String(svg_icons.TWITTER)

    @property
    def href(
        self,
    ):
        return f"https://twitter.com/intent/tweet?url={self._url_parsed}&text={self.text}"


class ShareOnReddit(ShareOnBase):
    """A Share on Reddit button"""

    icon = param.String(svg_icons.REDDIT)

    @property
    def href(
        self,
    ):
        return f"https://reddit.com/submit?url={self._url_parsed}&amp;title={self.text}"


class ShareOnMail(ShareOnBase):
    """A Share on Mail button"""

    icon = param.String(svg_icons.ENVELOPE)

    @property
    def href(
        self,
    ):
        return f"mailto:?subject={self._url_parsed}&amp;body={self.text}&nbsp;{self._url_parsed}"



"""Contains a method to get a logger"""
import logging

LEVEL = logging.INFO
LEVEL_NAME = "INFO"
# pylint: disable=implicit-str-concat
LOG_FORMAT = (
    "%(asctime)s :: %(levelname)s ::  %(filename)s :: %(funcName)s :: " "%(lineno)s :: %(message)s"
)

# file_handler = logging.FileHandler(filename='test.log', mode='w')
# file_handler.setFormatter(logging.Formatter(LOG_FORMAT))
# CONSOLE_HANDLER = logging.StreamHandler(sys.stdout)
# CONSOLE_HANDLER.setFormatter(logging.Formatter(LOG_FORMAT))


def get_logger(name, level=LEVEL):
    """Returns a logger"""
    logger = logging.getLogger(name)
    # logger.addHandler(CONSOLE_HANDLER)
    logger.setLevel(level)
    return logger


def log_started_message():
    """Starts the logger"""
    _logger = get_logger(__name__)
    _logger.info("Session Started")
    _logger.info("Log Level %s", LEVEL_NAME)



"""In this module we define the BaseModel.

The BaseModel adds ordering by the name parameter to a Class"""
import pathlib
from typing import Callable, Dict

import param
import toml


class BaseModel(param.Parameterized):
    """The BaseModel adds ordering by the name parameter to a Class"""

    uid = param.String(doc="A unique id identifying the model.")

    def __lt__(self, other):
        if hasattr(other, "name"):
            return self.name.casefold() < other.name.casefold()
        return True

    def __eq__(self, other):
        if hasattr(other, "name"):
            return self.name == other.name
        return False

    def __str__(
        self,
    ):
        return self.name

    def __repr__(
        self,
    ):
        return self.name

    @classmethod
    def create_from_toml(cls, path: pathlib.Path, clean_func: Callable = None) -> Dict:
        """Returns a list of Models from the toml file specified by the path"""
        config = toml.load(path)
        if not clean_func:
            clean_func = lambda x: x
        return {key: cls(uid=key, **clean_func(value)) for key, value in config.items()}



"""The FastGalleryTemplate can be used to show case your applications in a nice way"""
import pathlib

import param
from panel import Template

ROOT = pathlib.Path(__file__).parent
CSS = (ROOT / "fast_gallery_template.css").read_text()
JS = (ROOT / "fast_gallery_template.js").read_text()
TEMPLATE = (ROOT / "fast_gallery_template.html").read_text()
FAVICON = (
    "https://raw.githubusercontent.com/MarcSkovMadsen/awesome-panel/"
    "2781d86d4ed141889d633748879a120d7d8e777a/assets/images/favicon.ico"
)


class FastGalleryTemplate(Template):
    """The FastGalleryTemplate can be used to show case your applications in a nice way"""

    resources = param.List()
    title = param.String(
        default="Panel Application",
        doc="""
        A title to show in the header. Also added to the document head
        meta settings and as the browser tab title.""",
    )

    site = param.String(
        default="",
        doc="""
        The name of the site. Will be shown in the header and link to the
        root of the site. Default is '', i.e. not shown.""",
    )

    description = param.String("")
    background_image_url = param.String(
        "https://preview.redd.it/9oi428ohy7t21.png?auto=webp&s=5051b77d33e85446b6492a1e02725c6729777d4f"  # pylint: disable=line-too-long
    )
    target = param.ObjectSelector("_self", objects=["_blank", "_parent", "_top", "_self"])
    favicon = param.String(FAVICON)
    accent_base_color = param.String("#E1477E")
    font_family = param.String("Open Sans")
    font_url = param.String("//fonts.googleapis.com/css?family=Open+Sans")

    def __init__(self, **params):  # pylint: disable=too-many-arguments
        """The FastGalleryTemplate can be used to show case your applications in a nice way"""
        super().__init__(template=TEMPLATE, **params)
        self.add_variable("title_names", self.site + "|" + self.title)
        self.add_variable("site_name", self.site)
        self.add_variable("site_url", "/")
        self.add_variable("name", self.title)
        self.add_variable("url", "")
        self.add_variable("description", self.description)
        self.add_variable("background_image_url", self.background_image_url)
        self.add_variable("items", self.resources)
        self.add_variable("gallery_js", JS)
        self.add_variable("gallery_css", CSS)
        self.add_variable("favicon", self.favicon)
        self.add_variable("accent_base_color", self.accent_base_color)
        self.add_variable("font_family", self.font_family)
        self.add_variable("font_url", self.font_url)
        self.add_variable("target", self.target)



"""Implementation of MWC Material Components"""
import param

from awesome_panel_extensions.frameworks.material.config import MWC_ICONS
from awesome_panel_extensions.web_component import WebComponent

# pylint: disable=abstract-method


class Select(WebComponent):
    """Implementation of the mwc-select component

    The `value` is the value selected by the user. Can be None.
    The `options` are the options that can be selected by the user

    Set `outlined` to change the style
    """

    options = param.ClassSelector(
        default=[],
        class_=(dict, list),
        doc="A list or dictionary of options to select from options",
    )
    value = param.Parameter(doc="The current value; must be one of the option values")
    selects = param.Integer()
    _index = param.String()

    disabled = param.Boolean(
        default=False,
        doc="Whether or not the widget is editable",
    )

    icon = param.ObjectSelector(
        default=None,
        objects=MWC_ICONS,
        allow_None=True,
        doc="""Leading icon to display in select.""",
    )
    outlined = param.Boolean(
        default=False, doc="Whether or not to show the material outlined variant."
    )

    html = param.String("""<mwc-select style="width:100%"></mwc-select>""")
    attributes_to_watch = param.Dict(
        {"label": "name", "outlined": "outlined", "disabled": "disabled", "icon": "icon"}
    )
    properties_to_watch = param.Dict({"value": "_index"})
    events_to_watch = param.Dict(default={"selected": "selects"})
    parameters_to_watch = param.List(["options"])

    def __init__(self, min_height=60, **params):
        super().__init__(min_height=min_height, **params)

        self._set_class_()

    def _get_html_from_parameters_to_watch(self, **params) -> str:
        options = params["options"]
        if not options:
            return """<mwc-select></mwc-select>"""

        innerhtml = []
        if isinstance(options, list):
            for index, obj in enumerate(options):
                if hasattr(obj, "name"):
                    value = obj.name
                else:
                    value = str(obj)
                item = f'<mwc-list-item value="{index}">{value}</mwc-list-item>'
                innerhtml.append(item)
        if isinstance(options, dict):
            for index, value in enumerate(options.values()):
                item = f'<mwc-list-item value="{index}">{str(value)}</mwc-list-item>'
                innerhtml.append(item)

        return f"""<mwc-select>{"".join(innerhtml)}</mwc-select>"""

    @param.depends("options", watch=True)
    def _set_class_(self):
        if isinstance(self.options, list):
            self.param.options.class_ = list
        if isinstance(self.options, dict):
            self.param.options.class_ = dict

    @param.depends("value", watch=True)
    def _update_index(self):
        # pylint: disable=unsupported-membership-test
        if isinstance(self.options, list) and self.value in self.options:
            self._index = str(self.options.index(self.value))
        elif isinstance(self.options, dict) and self.value in self.options:
            self._index = str(list(self.options).index(self.value))
        else:
            self._index = ""

    @param.depends("_index", watch=True)
    def _update_value(self):
        if self._index == "":
            self.value = None
        elif isinstance(self.options, list):
            self.value = self.options[int(self._index)]  # pylint: disable=unsubscriptable-object
        elif isinstance(self.options, dict):
            self.value = list(self.options)[int(self._index)]
        else:
            self.value = None



"""Functionality for styling according to Fast.design"""
import pathlib
from typing import Dict

import param

_ROOT = pathlib.Path(__file__).parent.parent / "assets/css"
_CSS_FILES = [
    _ROOT / "fast_bokeh.css",
    _ROOT / "fast_bokeh_slickgrid.css",
    _ROOT / "fast_panel.css",
    _ROOT / "fast_panel_dataframe.css",
    _ROOT / "fast_panel_widgets.css",
    _ROOT / "fast_panel_markdown.css",
    _ROOT / "fast_awesome.css",
]
_ROOT_FILE = _ROOT / "fast_root.css"
_DEFAULT_ROOT_FILE = _ROOT / "fast_root_default.css"
_DARK_ROOT_FILE = _ROOT / "fast_root_dark.css"

COLLAPSED_SVG_ICON = """
<svg style="stroke: #E62F63" width="18" height="18" viewBox="0 0 18 18" fill="none" xmlns="http://www.w3.org/2000/svg" slot="collapsed-icon">
            <path d="M15.2222 1H2.77778C1.79594 1 1 1.79594 1 2.77778V15.2222C1 16.2041 1.79594 17 2.77778 17H15.2222C16.2041 17 17 16.2041 17 15.2222V2.77778C17 1.79594 16.2041 1 15.2222 1Z" stroke-linecap="round" stroke-linejoin="round"></path>
            <path d="M9 5.44446V12.5556" stroke-linecap="round" stroke-linejoin="round"></path>
            <path d="M5.44446 9H12.5556" stroke-linecap="round" stroke-linejoin="round"></path>
        </svg>
""".replace(
    "\n", ""
)

EXPANDED_SVG_ICON = """
<svg style="stroke: #E62F63" width="18" height="18" viewBox="0 0 18 18" fill="none" xmlns="http://www.w3.org/2000/svg" slot="expanded-icon">
    <path d="M15.2222 1H2.77778C1.79594 1 1 1.79594 1 2.77778V15.2222C1 16.2041 1.79594 17 2.77778 17H15.2222C16.2041 17 17 16.2041 17 15.2222V2.77778C17 1.79594 16.2041 1 15.2222 1Z" stroke-linecap="round" stroke-linejoin="round"></path>
    <path d="M5.44446 9H12.5556" stroke-linecap="round" stroke-linejoin="round"></path>
</svg>
""".replace(
    "\n", ""
)


def read_fast_css(
    root_file: pathlib.Path = _ROOT_FILE, style_root_file: pathlib.Path = _DEFAULT_ROOT_FILE
) -> str:
    """Returns the Fast Base CSS

    Args:
        root_file (pathlib.Path): A Path to a css file used across styles
        style_root_file (pathlib.Path): A Path to a css file used for a specific style
    Returns:
        str: [description]
    """
    css_files = [root_file, *_CSS_FILES, style_root_file]
    css_list = [file.read_text() for file in css_files]
    return "\n".join(css_list)


DEFAULT_CSS = read_fast_css(_ROOT_FILE, _DEFAULT_ROOT_FILE)
DARK_CSS = read_fast_css(_ROOT_FILE, _DARK_ROOT_FILE)
FONT_URL = "//fonts.googleapis.com/css?family=Open+Sans"


class FastStyle(param.Parameterized):
    """The FastStyle class provides the different colors and icons used to style the Fast
    Templates"""

    accent_fill_active = param.Color(default="#E45A8C")
    accent_fill_hover = param.Color(default="#DF3874")
    accent_fill_rest = param.Color(default="#A01346")
    accent_foreground_active = param.Color(default="#BA1651")
    accent_foreground_cut = param.Color(default="#000000")
    accent_foreground_hover = param.Color(default="#7A0F35")
    accent_foreground_rest = param.Color(default="#A01346")

    neutral_outline_active = param.Color(default="#D6D6D6")
    neutral_outline_hover = param.Color(default="#979797")
    neutral_outline_rest = param.Color(default="#BEBEBE")

    accent_base_color = param.Color(default="#A01346")
    background_color = param.Color(default="#ffffff")
    collapsed_icon = param.String(default=COLLAPSED_SVG_ICON)
    expanded_icon = param.String(default=EXPANDED_SVG_ICON)
    color = param.Color(default="#00aa41")
    neutral_fill_card_rest = param.Color(default="#F7F7F7")
    neutral_focus = param.Color(default="#888888")
    neutral_foreground_rest = param.Color(default="#2B2B2B")

    header_background = param.Color(default="#1B5E20")
    header_color = param.Color(default="#ffffff")
    css = param.String(default=DEFAULT_CSS)
    font = param.String(default="Open Sans, sans-serif")
    font_url = param.String(default=FONT_URL)

    def create_bokeh_theme(self) -> Dict:
        """Returns a custom bokeh theme based on the style parameters

        Returns:
            Dict: A Bokeh Theme
        """

        return {
            "attrs": {
                "Figure": {
                    "background_fill_color": self.background_color,
                    "border_fill_color": self.neutral_fill_card_rest,
                    "border_fill_alpha": 0,
                    "outline_line_color": self.neutral_focus,
                    "outline_line_alpha": 0.5,
                    "outline_line_width": 1,
                },
                "Grid": {"grid_line_color": self.neutral_focus, "grid_line_alpha": 0.25},
                "Axis": {
                    "major_tick_line_alpha": 0,
                    "major_tick_line_color": self.neutral_foreground_rest,
                    "minor_tick_line_alpha": 0,
                    "minor_tick_line_color": self.neutral_foreground_rest,
                    "axis_line_alpha": 0,
                    "axis_line_color": self.neutral_foreground_rest,
                    "major_label_text_color": self.neutral_foreground_rest,
                    "major_label_text_font": self.font,
                    "major_label_text_font_size": "1.025em",
                    "axis_label_standoff": 10,
                    "axis_label_text_color": self.neutral_foreground_rest,
                    "axis_label_text_font": self.font,
                    "axis_label_text_font_size": "1.25em",
                    "axis_label_text_font_style": "normal",
                },
                "Legend": {
                    "spacing": 8,
                    "glyph_width": 15,
                    "label_standoff": 8,
                    "label_text_color": self.neutral_foreground_rest,
                    "label_text_font": self.font,
                    "label_text_font_size": "1.025em",
                    "border_line_alpha": 0.5,
                    "border_line_color": self.neutral_focus,
                    "background_fill_alpha": 0.25,
                    "background_fill_color": self.neutral_fill_card_rest,
                },
                "ColorBar": {
                    "title_text_color": self.neutral_foreground_rest,
                    "title_text_font": self.font,
                    "title_text_font_size": "1.025em",
                    "title_text_font_style": "normal",
                    "major_label_text_color": self.neutral_foreground_rest,
                    "major_label_text_font": self.font,
                    "major_label_text_font_size": "1.025em",
                    # "background_fill_color": FAST_DARK_75,
                    "major_tick_line_alpha": 0,
                    "bar_line_alpha": 0,
                },
                "Title": {
                    "text_color": self.neutral_foreground_rest,
                    "text_font": self.font,
                    "text_font_size": "1.15em",
                },
            }
        }


DEFAULT_STYLE = FastStyle()
DARK_STYLE = FastStyle(
    header_background="#313131",
    header_color="#ffffff",
    accent_fill_active="#DC2567",
    accent_fill_hover="#E1477E",
    accent_fill_rest="#C01754",
    accent_foreground_active="#DF3874",
    accent_foreground_cut="#000000",
    accent_foreground_hover="#E55E8E",
    accent_foreground_rest="#E1477E",
    neutral_outline_active="#424242",
    neutral_outline_hover="#808080",
    neutral_outline_rest="#5A5A5A",
    accent_base_color="#E1477E",
    background_color="#181818",
    color="#ffffff",
    neutral_fill_card_rest="#212121",
    neutral_focus="#717171",
    neutral_foreground_rest="#e5e5e5",
    css=DARK_CSS,
)
DEFAULT_BOKEH_THEME = DEFAULT_STYLE.create_bokeh_theme()
DARK_BOKEH_THEME = DARK_STYLE.create_bokeh_theme()



"""The FastCheckBoxGroup enables using the `fast-checkbox` with Bokeh."""
from bokeh.core.properties import Bool
from bokeh.models.widgets import CheckboxGroup as _BkCheckboxGroup


class FastCheckboxGroup(_BkCheckboxGroup):
    """The FastCheckBoxGroup enables using the `fast-checkbox` with Bokeh."""

    readonly = Bool()



"""This module provides functionality for using the DataFrame widget like for example
custom styling"""
from typing import Dict

import pandas as pd
from bokeh.models.widgets.tables import NumberFormatter

INT_DTYPE = "int64"
INT_FORMAT = "0,0"
INT_ALIGN = "right"
FLOAT_DTYPE = "float64"
FLOAT_FORMAT = "0,0.00"
FLOAT_ALIGN = "right"


def get_default_formatters(
    data: pd.DataFrame,
    int_format: str = INT_FORMAT,
    int_align: str = INT_ALIGN,
    float_format: str = FLOAT_FORMAT,
    float_align: str = FLOAT_ALIGN,
) -> Dict:
    """A dictionary of columns and formats for the the `DataFrame` Panel widget.

    #### Example Use Case

    ```python
    formatters = get_default_formatters(data)
    dataframe_widget = pn.widgets.DataFrame(data, formatters=formatters))
    ```

    ### Notes

    - if the index is 1-dimensional and the name is None we rename to 'index' in order to be able
    to format the index.
    - For the complete formattings specification see [numbrojs](http://numbrojs.com/format.html)

    Args:
        data (pd.DataFrame): A DataFrame of data
        int_format (str, optional): The int format string. Defaults to INT_FORMAT.
        int_align (str, optional): [description]. Defaults to INT_ALIGN.
        float_format (str, optional): [description]. Defaults to FLOAT_FORMAT.
        float_align (str, optional): [description]. Defaults to FLOAT_ALIGN.

    Returns:
        Dict[str,str] -- A dictionary of default formats for the columns
    """
    formatters = {}
    for column in data.columns:
        if data[column].dtype == INT_DTYPE:
            formatters[column] = NumberFormatter(
                format=int_format,
                text_align=int_align,
            )
        elif data[column].dtype == FLOAT_DTYPE:
            formatters[column] = NumberFormatter(
                format=float_format,
                text_align=float_align,
            )

    # Hack: Without a proper name we cannot format the index
    if len(data.index.names) == 1 and not data.index.name:
        data.index.name = "index"

    for (
        index,
        name,
    ) in enumerate(data.index.names):
        if name:
            dtype = data.index.get_level_values(index).dtype
            if dtype == INT_DTYPE:
                formatters[name] = NumberFormatter(
                    format=int_format,
                    text_align=int_align,
                )
            elif dtype == FLOAT_DTYPE:
                formatters[name] = NumberFormatter(
                    format=float_format,
                    text_align=float_align,
                )

    return formatters



"""This module contains functionality to make any panel, i.e. layout, pane or widget, look like it's
loading and disabled.

This is very important for the user experience. A user that sees an unresponsive page without any
indication of something happening in the background with either perceive that the page is slow or
be in doubt if the page has stopped working.

THIS IS A COPY OF THE PANEL PR https://github.com/holoviz/panel/pull/1730
"""
_LOADING_INDICATOR_CSS_CLASS = "pn-loading"

# pylint: disable=line-too-long
DEFAULT_URL = "https://raw.githubusercontent.com/holoviz/panel/5ea166fdda6e1f958d2d9929ae2ed2b8e962156c/panel/assets/spinner_default.svg"
DARK_URL = "https://raw.githubusercontent.com/holoviz/panel/5ea166fdda6e1f958d2d9929ae2ed2b8e962156c/panel/assets/spinner_dark.svg"
# pylint: enable=line-too-long

STYLE = """
.bk.pn-loading:before {
    position: absolute;
    height: 100%;
    width: 100%;
    content: '';
    z-index: 1000;
    background-color: rgb(255,255,255,0.50);
    border-color: lightgray;
    background-image: url('https://raw.githubusercontent.com/MarcSkovMadsen/awesome-panel-assets/e6cb56375bb1c436975e09739a231fb31e628a63/spinners/default.svg');
    background-repeat: no-repeat;
    background-position: center;
    background-size: auto 50%;
    border-width: 1px;

}
.bk.pn-loading:hover:before {
    cursor: progress;
}
"""


def _add_css_class(*objects, css_class):
    for item in objects:
        if hasattr(item, "css_classes"):
            if not item.css_classes:
                item.css_classes = [css_class]
            elif not css_class in item.css_classes:
                item.css_classes.append(css_class)


def _remove_css_class(*objects, css_class):
    for item in objects:
        if hasattr(item, "css_classes"):
            if item.css_classes:
                if item.css_classes == [css_class]:
                    item.css_classes = None  # Will not work if set to []!
                elif css_class in item.css_classes:
                    item.css_classes.remove(css_class)


def start_loading_spinner(*objects):
    """
    Changes the appearance of the specified panel objects to indicate that they are loading.

    This is done by

    - adding a small spinner on top
    - graying out the panel
    - disabling the panel
    - and changing the mouse cursor to a spinner when hovering over the panel

    Arguments
    ---------
    objects: The panels to indicate are loading
    """
    _add_css_class(*objects, css_class=_LOADING_INDICATOR_CSS_CLASS)


def stop_loading_spinner(*objects):
    """Removes the loading indicating from the specified panel objects"""
    _remove_css_class(*objects, css_class=_LOADING_INDICATOR_CSS_CLASS)



"""A Model of an Author"""
from typing import List

import panel as pn
import param

from .base_model import BaseModel

STYLE = "img.pnx-avatar {height:100%;width:100%;border-radius:50%;vertical-align: middle;}"
if STYLE not in pn.config.raw_css:
    pn.config.raw_css.append(STYLE)


class Author(BaseModel):
    """A Model of an Author"""

    name = param.String(doc="The name of the person.")
    url = param.String(doc="A link to a page about the author.")
    avatar_url = param.String(doc="A link to an avatar image of the author.")

    twitter_url = param.String(doc="A link to the Twitter page of the Author")
    linkedin_url = param.String(doc="A link to the Linked In page of the Author")
    github_url = param.String(doc="A link to the Github page of the Author")

    all: List["Author"] = []

    def __init__(self, **params):
        super().__init__(**params)

        self.all.append(self)

    def _repr_html_(self):
        # pylint: disable=line-too-long
        return f"""<a href="{ self.url }" target="_blank"><img src="{ self.avatar_url }" class="pnx-avatar" alt="Avatar" title="{ self.name}"></a>"""



"""Provides the Application and Site"""
from . import category
from .application import Application
from .author import Author
from .resource import Resource
from .site import Site
from .site_config import SiteConfig



"""The FastButton enables using the `fast-button` with Bokeh."""
from bokeh.core import properties
from bokeh.models import Button as _BkButton


class FastButton(_BkButton):
    """The FastButton enables using the `fast-button` with Bokeh."""

    appearance = properties.String(
        default="neutral",
        help="The appearance attribute",
    )
    autofocus = properties.Bool(
        default=False,
        help="The autofocus attribute",
    )



"""The purpose of this module is to provide functionality for defining meta data about an
application. This can be done via the Application class"""
from typing import List

import param

from awesome_panel_extensions.site import category
from awesome_panel_extensions.site.resource import Resource


class Application(Resource):
    """The Application Class is a placeholder for Meta Data about an application"""

    category = param.String(default=category.APPLICATION)
    view = param.Parameter()
    servable = param.Parameter(
        doc="""A path to a .py file, a path to a .ipynb file or a callable
    returning the object to be served"""
    )

    all: List["Application"] = []

    def __init__(self, **params):
        super().__init__(**params)

        Resource.all.append(self)



"""List of Resource Categories"""
APPLICATION = "Application"
PERSON = "Person"
CODE = "Code"
IMAGE = "Image"
NOT_AVAILABLE = ""
OTHER = "Other"
TUTORIAL = "Tutorial"

ALL = [
    APPLICATION,
    PERSON,
    CODE,
    IMAGE,
    NOT_AVAILABLE,
    OTHER,
    TUTORIAL,
]



"""\
## Model Viewer

Google has developed the `model-viewer` web component for interactively viewing very large and
detailed 3D models.
"""
import panel as pn
import param

from awesome_panel_extensions.web_component import WebComponent

# pylint: disable=line-too-long
JS = """
<script src="https://unpkg.com/@webcomponents/webcomponentsjs@2.2.7/webcomponents-loader.js"></script>
<script type="module" src="https://unpkg.com/@google/model-viewer/dist/model-viewer.js"></script>
<script nomodule src="https://unpkg.com/@google/model-viewer/dist/model-viewer-legacy.js"></script>
<script src="https://unpkg.com/resize-observer-polyfill@1.5.1/dist/ResizeObserver.js"></script>
"""

HTML = """
<model-viewer src="https://modelviewer.dev/shared-assets/models/Astronaut.glb" alt="A 3D model of an astronaut"
auto-rotate camera-controls>
</model-viewer>
"""

MODELS = {
    "Astronaut": "https://modelviewer.dev/shared-assets/models/Astronaut.glb",
    "Boom Box": "https://modelviewer.dev/shared-assets/models/glTF-Sample-Models/2.0/BoomBox/glTF-Binary/BoomBox.glb",
    "Brain Stem": "https://modelviewer.dev/shared-assets/models/glTF-Sample-Models/2.0/BrainStem/glTF-Binary/BrainStem.glb",
    "Corset": "https://modelviewer.dev/shared-assets/models/glTF-Sample-Models/2.0/Corset/glTF-Binary/Corset.glb",
    "Damaged Helmet": "https://modelviewer.dev/shared-assets/models/glTF-Sample-Models/2.0/DamagedHelmet/glTF-Binary/DamagedHelmet.glb",
    "Flight Helmet": "https://modelviewer.dev/shared-assets/models/glTF-Sample-Models/2.0/FlightHelmet/glTF/FlightHelmet.gltf",
    "Lantern": "https://modelviewer.dev/shared-assets/models/glTF-Sample-Models/2.0/Lantern/glTF-Binary/Lantern.glb",
    "Monkey": "https://modelviewer.dev/shared-assets/models/glTF-Sample-Models/2.0/Suzanne/glTF/Suzanne.gltf",
    "Water Bottles": "https://modelviewer.dev/shared-assets/models/glTF-Sample-Models/2.0/SpecGlossVsMetalRough/glTF-Binary/SpecGlossVsMetalRough.glb",
    "Robot Expressive": "https://modelviewer.dev/shared-assets/models/RobotExpressive.glb",
    "Transparency Test": "https://modelviewer.dev/shared-assets/models/alpha-blend-litmus.glb",
    "Metal Rough Spheres": "https://modelviewer.dev/shared-assets/models/glTF-Sample-Models/2.0/MetalRoughSpheres/glTF/MetalRoughSpheres.gltf",
}
# pylint: disable=line-too-long

SRC_DEFAULT = MODELS["Flight Helmet"]

HEIGHT_DEFAULT = 600
HEIGHT_BOUNDS = (50, 1000)
WIDTH_DEFAULT = 600
WIDTH_BOUNDS = (50, 1000)

BACKGROUND = "#9E9E9E"

PARAMETERS = [
    "src",
    "height",
    "width",
    "exposure",
    # "auto_rotate",
    # "camera_controls",
]


class ModelViewer(WebComponent):
    """A Wired ModelViewer"""

    html = param.String(HTML)
    attributes_to_watch = param.Dict({"src": "src"})
    properties_to_watch = param.Dict(
        {
            "exposure": "exposure",
            "auto-rotate": "auto_rotate",
            "camera-controls": "camera_controls",
        }
    )

    src = param.ObjectSelector(default=SRC_DEFAULT, objects=MODELS)
    exposure = param.Number(1.0, bounds=(0, 2))
    auto_rotate = param.Boolean()
    camera_controls = param.Boolean()

    height = param.Integer(default=HEIGHT_DEFAULT, bounds=HEIGHT_BOUNDS)
    width = param.Integer(default=WIDTH_DEFAULT, bounds=WIDTH_BOUNDS)

    style = param.String()

    def __init__(self, **params):
        super().__init__(**params)

        self.css_pane = pn.pane.HTML()
        self.js_pane = pn.pane.HTML(JS)

        self._update_height_and_width()

    def view(self) -> pn.Column:
        """Returns the main view of the ModelViewer

        Returns:
            pn.Column: The main view of the ModelViewer
        """

        return pn.Column(
            self,
            self.js_pane,
            self.css_pane,
            sizing_mode="stretch_both",
        )

    @param.depends("height", "width", watch=True)
    def _update_height_and_width(self):
        if self.height:
            height = self.height
        else:
            height = HEIGHT_DEFAULT
        if self.width:
            width = self.width
        else:
            width = WIDTH_DEFAULT

        self.css_pane.object = f"""
<style>
model-viewer {{
    height:{height}px;
    width:{width}px;
}}
</style>
"""



"""Implementation of the Tabulator pane.

See http://tabulator.info/
"""

from bokeh.core import properties
from bokeh.models import ColumnDataSource
from bokeh.models.layouts import HTMLBox

# pylint: disable=line-too-long

JS_SRC = "https://unpkg.com/tabulator-tables@4.7.2/dist/js/tabulator.min.js"
MOMENT_SRC = "https://unpkg.com/moment@2.27.0/moment.js"

CSS_HREFS = {
    "default": "https://unpkg.com/tabulator-tables@4.7.2/dist/css/tabulator.min.css",
    "site": "https://unpkg.com/tabulator-tables@4.7.2/dist/css/tabulator_site.min.css",
    "simple": "https://unpkg.com/tabulator-tables@4.7.2/dist/css/tabulator_simple.min.css",
    "midnight": "https://unpkg.com/tabulator-tables@4.7.2/dist/css/tabulator_midnight.min.css",
    "modern": "https://unpkg.com/tabulator-tables@4.7.2/dist/css/tabulator_modern.min.css",
    "bootstrap": "https://unpkg.com/tabulator-tables@4.7.2/dist/css/bootstrap/tabulator_bootstrap.min.css",
    "bootstrap4": "https://unpkg.com/tabulator-tables@4.7.2/dist/css/bootstrap/tabulator_bootstrap4.min.css",
    "semantic-ui": "https://unpkg.com/tabulator-tables@4.7.2/dist/css/semantic-ui/tabulator_semantic-ui.min.css",
    "bulma": "https://unpkg.com/tabulator-tables@4.7.2/dist/css/bulma/tabulator_bulma.min.css",
    "materialize": "https://unpkg.com/tabulator-tables@4.7.2/dist/css/materialize/tabulator_materialize.min.css",
}

# pylint: enable=line-too-long


class TabulatorModel(HTMLBox):
    """A Bokeh Model that enables easy use of Tabulator tables

    See http://tabulator.info/
    """

    __javascript__ = [
        MOMENT_SRC,
        JS_SRC,
    ]

    # I could not get Tabulator loaded in Notebook
    # I found a working solution using requirejs the notebook
    # I'm working on getting the below working.
    # See https://github.com/holoviz/panel/issues/1529
    __js_skip__ = {"Tabulator": __javascript__[1:]}

    __js_require__ = {
        "paths": {"tabulator": ["https://unpkg.com/tabulator-tables@4.7.2/dist/js/tabulator.min"]},
        "exports": {"tabulator": "Tabulator"},
    }

    # __css__ = [CSS_HREFS["default"]]

    configuration = properties.Dict(properties.String, properties.Any)
    source = properties.Instance(ColumnDataSource)
    _cell_change = properties.Dict(properties.String, properties.Any)



"""Module containing functionality to work with colors. For example basic transformations between
formats like rgb, hex and hsl"""
# pylint: disable=invalid-name, missing-function-docstring
import colorsys
import math
from typing import Dict


# https://stackoverflow.com/questions/3380726/converting-a-rgb-color-tuple-to-a-six-digit-code-in-python
def rgb_to_hex(rgb):
    r = int(round(rgb[0]))
    g = int(round(rgb[1]))
    b = int(round(rgb[2]))
    return "#{:02x}{:02x}{:02x}".format(r, g, b)


def hex_to_rgb(color):
    color = color.replace("#", "")
    return tuple(int(color[i : i + 2], 16) for i in (0, 2, 4))


# https://gist.github.com/mathebox/e0805f72e7db3269ec22
def rgb_to_hsl(rgb):
    rgb = tuple(float(item) / 255 for item in rgb)
    h, l, s = colorsys.rgb_to_hls(*rgb)
    return (h, s, l)


def hsl_to_rgb(hsl):
    h, s, l = hsl
    r, g, b = colorsys.hls_to_rgb(h, l, s)
    r = int(round(r * 255))
    g = int(round(g * 255))
    b = int(round(b * 255))
    return (r, g, b)


def hex_to_hsl(color):
    rgb = hex_to_rgb(color)
    return rgb_to_hsl(rgb)


def hsl_to_hex(hsl):
    rgb = hsl_to_rgb(hsl)
    return rgb_to_hex(rgb)


# https://github.com/bgrins/TinyColor/blob/master/tinycolor.js
def get_brightness(color):
    rgb = hex_to_rgb(color)
    return (rgb[0] * 299 + rgb[1] * 587 + rgb[2] * 114) / 1000


def is_dark(color):
    return get_brightness(color) < 128


def is_light(color):
    return not is_dark(color)


# https://github.com/bgrins/TinyColor/blob/master/tinycolor.js
def mix(color1: str, color2: str, amount: float) -> str:
    if amount < 0:
        amount = 0
    elif amount > 100:
        amount = 100

    rgb1 = hex_to_rgb(color1)
    rgb2 = hex_to_rgb(color2)

    ratio = amount / 100

    rgb = (
        (rgb2[0] - rgb1[0]) * ratio + rgb1[0],
        (rgb2[1] - rgb1[1]) * ratio + rgb1[1],
        (rgb2[2] - rgb1[2]) * ratio + rgb1[2],
    )
    return rgb_to_hex(rgb)


# https://github.com/mbitson/mcg/blob/master/scripts/controllers/ColorGeneratorCtrl.js
def multiply(color1, color2):
    rgb1 = hex_to_rgb(color1)
    rgb2 = hex_to_rgb(color2)

    rgb = (
        math.floor(rgb1[0] * rgb2[0] / 255),
        math.floor(rgb1[1] * rgb2[1] / 255),
        math.floor(rgb1[2] * rgb2[2] / 255),
    )
    return rgb_to_hex(rgb)


def clamp01(val):
    if val < 0:
        return 0
    if val > 1:
        return 1
    return val


def saturate(color, amount):
    if not amount:
        amount = 0
    elif amount < 0:
        amount = 0
    elif amount > 100:
        amount = 100

    hsl = hex_to_hsl(color)
    h, s, l = hsl
    s += amount / 100
    s = clamp01(s)
    colorsat = hsl_to_hex((h, s, l))
    return colorsat


def lighten(color, amount):
    if not amount:
        amount = 0
    elif amount < 0:
        amount = 0
    elif amount > 100:
        amount = 100

    hsl = hex_to_hsl(color)
    h, s, l = hsl
    l += amount / 100
    l = clamp01(l)
    colorlight = hsl_to_hex((h, s, l))
    return colorlight


def tetrad(color):
    hsl = hex_to_hsl(color)
    h, s, l = hsl
    hsl1 = (((h + 90) % 360) / 360, s, l)
    hsl2 = (((h + 180) % 360) / 360, s, l)
    hsl3 = (((h + 270) % 360) / 360, s, l)

    return [
        color,
        hsl_to_hex(hsl1),
        hsl_to_hex(hsl2),
        hsl_to_hex(hsl3),
    ]


# https://github.com/mbitson/mcg/blob/master/scripts/controllers/ColorGeneratorCtrl.js
def compute_colors(color: str) -> Dict:
    baselight = "#ffffff"
    basedark = multiply(color, color)
    basetriad = tetrad(color)
    return {
        "50": mix(baselight, color, 12),
        "100": mix(baselight, color, 30),
        "200": mix(baselight, color, 50),
        "300": mix(baselight, color, 70),
        "400": mix(baselight, color, 85),
        "500": mix(baselight, color, 100),
        "600": mix(basedark, color, 87),
        "700": mix(basedark, color, 70),
        "800": mix(basedark, color, 54),
        "900": mix(basedark, color, 25),
        "A100": lighten(saturate(mix(basedark, basetriad[3], 15), 80), 65),
        "A200": lighten(saturate(mix(basedark, basetriad[3], 15), 80), 55),
        "A400": lighten(saturate(mix(basedark, basetriad[3], 15), 100), 45),
        "A700": lighten(saturate(mix(basedark, basetriad[3], 15), 100), 40),
    }



"""The FastTemplate provides a Panel Template with the Fast javascript library loaded."""
import param
from panel import Template

FAST_CSS = """
html {
    height:100%;
}
html, fast-design-system-provider {
    min-height: 100vh;
}
body {
    margin: 0px;
    padding: 0;
    font-style: normal;
    font-variant-ligatures: normal;
    font-variant-caps: normal;
    font-variant-numeric: normal;
    font-variant-east-asian: normal;
    font-weight: normal;
    font-stretch: normal;
    font-size: 16px;
    line-height: normal;
    font-family: aktiv-grotesk, "Segoe UI", Arial, Helvetica, sans-serif;
    overflow-x: hidden;
    overflow-y: hidden;
}
.bk-root .bk-fast-input {
  display: inline-block;
  width: 100%;
  flex-grow: 1;
  -webkit-flex-grow: 1;
  min-height: 31px;
  padding: 0 12px;
}
"""
# The key to the template is to insert the '<fast-design-system-provider>' tag around the content
# For now the FAST_CSS is included directly in the template. Will have to find a better way.
# Please also note that the template loads the fast js dependency as Panel does not currently
# support loading js modules.
# Please also note that for usage on server the height is set to 100vh to not get a mix of
# black and white backgrounds. We only want the background from the fast-design-system-provider
TEMPLATE = """
{% from macros import embed %}

<!DOCTYPE html>
<html lang="en">
{% block head %}
<head>
    {% block inner_head %}
    <meta charset="utf-8">
    <title>{% block title %}{{ title | e if title else "Panel App" }}{% endblock %}</title>
    {% block preamble %}{% endblock %}
    {% block resources %}
        {% block css_resources %}
        {{ bokeh_css | indent(8) if bokeh_css }}
        {% endblock %}
        {% block js_resources %}
        {{ bokeh_js | indent(8) if bokeh_js }}
        {% endblock %}
    {% endblock %}
    {% block postamble %}
        <link href="https://use.typekit.net/spx2dgm.css" rel="stylesheet">
        <style>
            html {
                height:100%;
            }
            html, fast-design-system-provider {
                min-height: 100vh;
            }
            body {
                margin: 0px;
                padding: 0;
                font-style: normal;
                font-variant-ligatures: normal;
                font-variant-caps: normal;
                font-variant-numeric: normal;
                font-variant-east-asian: normal;
                font-weight: normal;
                font-stretch: normal;
                font-size: 16px;
                line-height: normal;
                font-family: aktiv-grotesk, "Segoe UI", Arial, Helvetica, sans-serif;
            }
            .bk-root .bk-fast-input {
                display: inline-block;
                width: 100%;
                flex-grow: 1;
                -webkit-flex-grow: 1;
            }
        </style>
    {% endblock %}
    {% endblock %}
</head>
{% endblock %}
{% block body %}
<body>
    <fast-design-system-provider use-defaults>
        {% block inner_body %}
        {% block contents %}
            {% for doc in docs %}
            {{ embed(doc) if doc.elementid }}
            {% for root in doc.roots %}
                {{ embed(root) | indent(10) }}
            {% endfor %}
            {% endfor %}
        {% endblock %}
        {{ plot_script | indent(8) }}
        {% endblock %}
    </fast-design-system-provider>
    <script type="module" src="https://unpkg.com/@microsoft/fast-components"></script>
</body>
{% endblock %}
</html>
"""

NB_TEMPLATE = """
{% from macros import embed %}

<!DOCTYPE html>
<html lang="en">
{% block head %}
<head>
    {% block inner_head %}
    <meta charset="utf-8">
    <title>{% block title %}{{ title | e if title else "Panel App" }}{% endblock %}</title>
    {% block preamble %}{% endblock %}
    {% block resources %}
        {% block css_resources %}
        {{ bokeh_css | indent(8) if bokeh_css }}
        {% endblock %}
        {% block js_resources %}
        {{ bokeh_js | indent(8) if bokeh_js }}
        {% endblock %}
    {% endblock %}
    {% block postamble %}
        <style>
            body {
                margin: 0px;
                padding: 0;
                font-style: normal;
                font-variant-ligatures: normal;
                font-variant-caps: normal;
                font-variant-numeric: normal;
                font-variant-east-asian: normal;
                font-weight: normal;
                font-stretch: normal;
                font-size: 16px;
                line-height: normal;
                font-family: aktiv-grotesk, "Segoe UI", Arial, Helvetica, sans-serif;
            }
            .bk-root .bk-fast-input {
                display: inline-block;
                width: 100%;
                flex-grow: 1;
                -webkit-flex-grow: 1;
            }
        </style>
    {% endblock %}
    {% endblock %}
</head>
{% endblock %}
{% block body %}
<body>
    <fast-design-system-provider use-defaults>
        {% block inner_body %}
        {% block contents %}
            {% for doc in docs %}
            {{ embed(doc) if doc.elementid }}
            {% for root in doc.roots %}
                {{ embed(root) | indent(10) }}
            {% endfor %}
            {% endfor %}
        {% endblock %}
        {{ plot_script | indent(8) }}
        {% endblock %}
    </fast-design-system-provider>
    <script type="module" src="https://unpkg.com/@microsoft/fast-components"></script>
</body>
{% endblock %}
</html>
"""


class FastTemplate(Template):
    """The FastTemplate provides a Panel Template with the Fast javascript library loaded."""

    main = param.ClassSelector(
        class_=list,
        constant=True,
        doc="""
        A list-like container which populates the main area.""",
    )

    def __init__(self, main):
        if not isinstance(main, list):
            main = [main]
        items = {str(key): value for key, value in enumerate(main)}
        super().__init__(template=TEMPLATE, nb_template=NB_TEMPLATE, items=items, main=main)



"""The FastListTemplate provides a list layout based on similar to the
Panel VanillaTemplate but in the Fast.design style and enabling the
use of Fast components.
"""
import pathlib

import param
from bokeh.themes import Theme as _BkTheme
from panel.template.theme import DarkTheme, DefaultTheme

from awesome_panel_extensions.frameworks.fast import styles
from awesome_panel_extensions.frameworks.fast.templates.base import BasicTemplate


class FastListTemplate(BasicTemplate):
    """
    The FastListTemplate is build on top of Fast.design.
    """

    _css = pathlib.Path(__file__).parent / "fast_list_template.css"
    _js = pathlib.Path(__file__).parent.parent.parent / "assets/js/fast_template.js"

    _template = pathlib.Path(__file__).parent / "fast_list_template.html"

    def _get_theme(self, name: str = "default"):
        if name == "dark":
            return FastDarkTheme
        return FastDefaultTheme


class FastDefaultTheme(DefaultTheme):
    """The Default Theme of the FastListTemplate"""

    css = param.Filename(default=pathlib.Path(__file__).parent / "default.css")

    _template = FastListTemplate

    style = param.ClassSelector(class_=styles.FastStyle, default=styles.DEFAULT_STYLE)

    bokeh_theme = param.ClassSelector(
        class_=(_BkTheme, str), default=_BkTheme(json=styles.DEFAULT_BOKEH_THEME)
    )


class FastDarkTheme(DarkTheme):
    """The Dark Theme of the FastListTemplate"""

    css = param.Filename(default=pathlib.Path(__file__).parent / "dark.css")

    _template = FastListTemplate

    style = param.ClassSelector(class_=styles.FastStyle, default=styles.DARK_STYLE)

    bokeh_theme = param.ClassSelector(
        class_=(_BkTheme, str), default=_BkTheme(json=styles.DARK_BOKEH_THEME)
    )
